<?php return array(
    'root' => array(
        'name' => '__root__',
        'pretty_version' => 'dev-main',
        'version' => 'dev-main',
        'reference' => 'd4d9bba3c5904cdebea5c9bebab27d8abd6c1af1',
        'type' => 'library',
        'install_path' => __DIR__ . '/../../',
        'aliases' => array(),
        'dev' => true,
    ),
    'versions' => array(
        '__root__' => array(
            'pretty_version' => 'dev-main',
            'version' => 'dev-main',
            'reference' => 'd4d9bba3c5904cdebea5c9bebab27d8abd6c1af1',
            'type' => 'library',
            'install_path' => __DIR__ . '/../../',
            'aliases' => array(),
            'dev_requirement' => false,
        ),
        'ergebnis/phpstan-rules' => array(
            'pretty_version' => '2.12.0',
            'version' => '2.12.0.0',
            'reference' => 'c4e0121a937b3b551f800a86e7d78794da2783ea',
            'type' => 'phpstan-extension',
            'install_path' => __DIR__ . '/../ergebnis/phpstan-rules',
            'aliases' => array(),
            'dev_requirement' => true,
        ),
        'nette/utils' => array(
            'pretty_version' => 'v4.1.1',
            'version' => '4.1.1.0',
            'reference' => 'c99059c0315591f1a0db7ad6002000288ab8dc72',
            'type' => 'library',
            'install_path' => __DIR__ . '/../nette/utils',
            'aliases' => array(),
            'dev_requirement' => true,
        ),
        'phpstan/extension-installer' => array(
            'pretty_version' => '1.4.3',
            'version' => '1.4.3.0',
            'reference' => '85e90b3942d06b2326fba0403ec24fe912372936',
            'type' => 'composer-plugin',
            'install_path' => __DIR__ . '/../phpstan/extension-installer',
            'aliases' => array(),
            'dev_requirement' => true,
        ),
        'phpstan/phpstan' => array(
            'pretty_version' => '2.1.33',
            'version' => '2.1.33.0',
            'reference' => '9e800e6bee7d5bd02784d4c6069b48032d16224f',
            'type' => 'library',
            'install_path' => __DIR__ . '/../phpstan/phpstan',
            'aliases' => array(),
            'dev_requirement' => true,
        ),
        'phpstan/phpstan-strict-rules' => array(
            'pretty_version' => '2.0.7',
            'version' => '2.0.7.0',
            'reference' => 'd6211c46213d4181054b3d77b10a5c5cb0d59538',
            'type' => 'phpstan-extension',
            'install_path' => __DIR__ . '/../phpstan/phpstan-strict-rules',
            'aliases' => array(),
            'dev_requirement' => true,
        ),
        'tomasvotruba/type-coverage' => array(
            'pretty_version' => '2.1.0',
            'version' => '2.1.0.0',
            'reference' => '468354b3964120806a69890cbeb3fcf005876391',
            'type' => 'phpstan-extension',
            'install_path' => __DIR__ . '/../tomasvotruba/type-coverage',
            'aliases' => array(),
            'dev_requirement' => true,
        ),
    ),
);
